using DataStructures

in(node::Node, graph::DAG) = node in graph.nodes
in(edge::Edge, graph::DAG) = edge in graph.edges

function is_parent(potential_parent, node)
   return potential_parent in node.parents
end

function is_child(potential_child, node)
   return potential_child in node.children
end

function ==(n1::Node, n2::Node, g::DAG)
   if typeof(n1) != typeof(n2)
      return false
   end
   if !(n1 in g) || !(n2 in g)
      return false
   end

   return n1.task == n2.task && children(n1) == children(n2)
end

# children = prerequisite nodes, nodes that need to execute before the task, edges point into this task
function children(node::Node)
   return copy(node.children)
end

# parents = subsequent nodes, nodes that need this node to execute, edges point from this task
function parents(node::Node)
   return copy(node.parents)
end

# siblings = all children of any parents, no duplicates, does not include the node itself
function siblings(node::Node)
   result = Set{Node}()
   for parent in parents(node)
      for sibling in children(parent)
         if (sibling != node)
            push!(result, sibling)
         end
      end
   end

   return result
end

# partners = all parents of any children, no duplicates, does not include the node itself
function partners(node::Node)
   result = Set{Node}()
   for child in children(node)
      for partner in parents(node)
         if (partner != node)
            push!(result, partner)
         end
      end
   end

   return result
end

is_entry_node(node::Node) = length(children(node)) == 0
is_exit_node(node::Node) = length(parents(node)) == 0

function insert_node(graph::DAG, node::Node)
   push!(graph.nodes, node)
   return node
end

function insert_edge(graph::DAG, edge::Edge)
   # edge points from child to parent
   push!(edge.edge[1].parents, edge.edge[2])
   push!(edge.edge[2].children, edge.edge[1])

   return edge
end

function remove_node(graph::DAG, node::Node)
   delete!(graph.nodes, node)
   return nothing
end

function remove_edge(graph::DAG, edge::Edge)
   filter!(x -> x != edge.edge[2], edge.edge[1].parents)
   filter!(x -> x != edge.edge[1], edge.edge[2].children)
   return nothing
end

function graph_properties(graph::DAG)
   d = 0
   ce = 0
   for node in graph.nodes
      d += data(node.task)
      ce += compute_effort(node.task)
   end

   ci = ce / d

   result = (data = d,
             compute_effort = ce,
             compute_intensity = ci)
   return result
end

function get_exit_node(graph::DAG)
   for node in graph.nodes
      if (is_exit_node(node))
         return node
      end
   end
   error("The given graph has no exit node! It is either empty or not acyclic!")
end

function can_reduce(n1::Node, n2::Node)
   if (n1.task != n2.task)
       return false
   end

   return parents(n1) == parents(n2)
end

function can_split(n::Node)
   return length(parents(n)) > 1
end

# check whether the given graph is connected
function is_valid(graph::DAG)
   nodeQueue = Deque{Node}()
   push!(nodeQueue, get_exit_node(graph))
   seenNodes = Set{Node}()

   while ! isempty(nodeQueue)
      current = pop!(nodeQueue)
      push!(seenNodes, current)

      childrenNodes = children(current)
      for child in childrenNodes
         push!(nodeQueue, child)
      end
   end

   return length(seenNodes) == length(graph.nodes)
end

function show_nodes(io, graph::DAG)
   print(io, "[")
   first = true
   for n in graph.nodes
      if first
         first = false
      else
         print(io, ", ")
      end
      print(io, n)
   end
   print(io, "]")
end

function show(io::IO, graph::DAG)
   println(io, "Graph:")
   print(io, "  Nodes: ")

   nodeDict = Dict{Type, Int64}()
   noEdges = 0
   for node in graph.nodes
      if haskey(nodeDict, typeof(node.task))
         nodeDict[typeof(node.task)] = nodeDict[typeof(node.task)] + 1
      else
         nodeDict[typeof(node.task)] = 1
      end
      noEdges += length(parents(node))
   end

   if length(graph.nodes) <= 20
      show_nodes(io, graph)
   else
      print("Total: ", length(graph.nodes), ", ")
      first = true
      for (type, number) in zip(keys(nodeDict), values(nodeDict))
         if first
            first = false
         else
            print(", ")
         end
         print(type, ": ", number)
      end
   end
   println(io)
   println(io, "  Edges: ", noEdges)
   properties = graph_properties(graph)
   println(io, "  Total Compute Effort: ", properties.compute_effort)
   println(io, "  Total Data Transfer: ", properties.data)
   println(io, "  Total Compute Intensity: ", properties.compute_intensity)
end
